unit convert;

interface

// converter function type
type
  Tconvert = procedure(src: pointer; dest: pointer; pixels: Integer);

// converter request
function ptc_request_converter(bits, r, g,b: integer): TConvert;


implementation

type
  PLongA = ^TLongA;
  TLongA = array [0..0] of Longint;
  PSmallA = ^TSmallA;
  TSmallA = array [0..0] of Smallint;
  PByteA = ^TByteA;
  TByteA = array [0..0] of Byte;


{$ifdef __PTC_MMX__}
procedure mmx_memcpy(src: pointer; dest: pointer; pixels: Integer);
asm
    push ebp
    mov ebp,esp

    pushad

    mov edi,dest        //; destination
    mov esi,src         //; source
    mov ecx,pixels      //; bytes

    mov eax,ecx
    shr ecx,6
    mov ebx,ecx
    shl ebx,6
    sub eax,ebx

@@loop:

      db $0F,$6F,$06           /// movq mm0,[esi]
      db $0F,$6F,$4E,$08       /// movq mm1,[esi+8]
      db $0F,$6F,$56,$10       /// movq mm2,[esi+16]
      db $0F,$6F,$5E,$18       /// movq mm3,[esi+24]
      db $0F,$6F,$66,$20       /// movq mm4,[esi+32]
      db $0F,$6F,$6E,$28       /// movq mm5,[esi+40]
      db $0F,$6F,$76,$30       /// movq mm6,[esi+48]
      db $0F,$6F,$7E,$38       /// movq mm7,[esi+56]
      db $0F,$7F,$07           /// movq [edi],mm0
      db $0F,$7F,$4F,$08       /// movq [edi+8],mm1
      db $0F,$7F,$57,$10       /// movq [edi+16],mm2
      db $0F,$7F,$5F,$18       /// movq [edi+24],mm3
      db $0F,$7F,$67,$20       /// movq [edi+32],mm4
      db $0F,$7F,$6F,$28       /// movq [edi+40],mm5
      db $0F,$7F,$77,$30       /// movq [edi+48],mm6
      db $0F,$7F,$7F,$38       /// movq [edi+56],mm7
      add esi, 64
      add edi, 64
      dec ecx
      jnz @@loop

    mov ecx,eax
    rep movsb

    db $0F,$77               /// emms

    popad

    pop ebp
    ret
end;
{$endif}


procedure ptc_convert_32_to_32_rgb888(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p, q: PLongA;
{$ifndef __PTC_MEMCPY__}
  i: integer;
{$endif}
{$endif}
begin
{$ifdef __PTC_MMX__}
  mmx_memcpy(src, dest, pixels shl 2);
{$else}
{$ifdef __PTC_MEMCPY__}
   p := src;
   q := dest;
   move(p[0], q[0], pixels shl 2);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
     q[0] := p[0];
     Inc(p);
     Inc(q);
   end;
{$endif}
{$endif}
end;

procedure ptc_convert_32_to_32_bgr888(src: pointer; dest: pointer; pixels: Integer);
var
  p, q: PLongA;
  i: integer;
  r,g,b: integer;
begin
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
     r := (p[0] and $00FF0000) shr 16;
     g := (p[0] and $0000FF00);
     b := (p[0] and $000000FF) shl 16;
     q[0] := r or g or b;
     Inc(p);
     Inc(q);
   end;
end;

procedure ptc_convert_32_to_24_rgb888(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p, q: PByteA;
  i: integer;
{$endif}
begin
{$ifdef __PTC_MMX__}
//   mmx_convert_32_to_24_rgb888(dst, src, pixels);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
{$ifdef __PTC_LITTLE_ENDIAN__}
     q[0] := p[0];
     Inc(p);
     Inc(q);
     q[0] := p[0];
     Inc(p);
     Inc(q);
     q[0] := p[0];
     Inc(p, 2);
     Inc(q);
{$else}
     Inc(q, 2);
     q[0] := p[0];
     Inc(p);
     Dec(q);
     q[0] := p[0];
     Inc(p);
     Dec(q);
     q[0] := p[0];
     Inc(p, 2);
     Inc(q, 3);
{$endif}
     Integer(p) := Integer(p) + 4;
     Integer(q) := Integer(q) + 3;
   end;
{$endif}
end;

procedure ptc_convert_32_to_24_bgr888(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p, q: PByteA;
  i: integer;
{$endif}
begin
{$ifdef __PTC_MMX__}
//   mmx_convert_32_to_24_bgr888(dst, src, pixels);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
{$ifdef __PTC_LITTLE_ENDIAN__}
     Inc(q, 2);
     q[0] := p[0];
     Inc(p);
     Dec(q);
     q[0] := p[0];
     Inc(p);
     Dec(q);
     q[0] := p[0];
     Inc(p, 2);
     Inc(q, 3);
{$else}
     q[0] := p[0];
     Inc(p);
     Inc(q);
     q[0] := p[0];
     Inc(p);
     Inc(q);
     q[0] := p[0];
     Inc(p, 2);
     Inc(q);
{$endif}
   end;
{$endif}
end;


procedure ptc_convert_32_to_16_rgb565(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p: PLongA;
  q: PSmallA;
  i: integer;
  r,g,b: integer;
{$endif}
begin
{$ifdef __PTC_MMX__}
//   mmx_convert_32_to_16_rgb565(dst, src, pixels);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
     r := (p[0] and $00F80000) shr 8;
     g := (p[0] and $0000FC00) shr 5;
     b := (p[0] and $000000F8) shr 3;
     q[0] := r or g or b;
     Inc(p);
     Inc(q);
   end;
{$endif}
end;

procedure ptc_convert_32_to_16_bgr565(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p: PLongA;
  q: PSmallA;
  i: integer;
  r,g,b: integer;
{$endif}
begin
{$ifdef __PTC_MMX__}
//   mmx_convert_32_to_16_bgr565(dst, src, pixels);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
     r := (p[0] and $00F80000) shr 19;
     g := (p[0] and $0000FC00) shr 5;
     b := (p[0] and $000000F8) shl 8;
     q[0] := r or g or b;
     Inc(p);
     Inc(q);
   end;
{$endif}
end;

procedure ptc_convert_32_to_16_rgb555(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p: PLongA;
  q: PSmallA;
  i: integer;
  r,g,b: smallInt;
{$endif}
begin
{$ifdef __PTC_MMX__}
//   mmx_convert_32_to_16_rgb555(dst, src, pixels);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
     r := (p[0] and $00F80000) shr 9;
     g := (p[0] and $0000F800) shr 6;
     b := (p[0] and $000000F8) shr 3;

     q[0] := (r or g or b);
     Inc(p);
     Inc(q);
   end;

{$endif}
end;

procedure ptc_convert_32_to_16_bgr555(src: pointer; dest: pointer; pixels: Integer);
{$ifndef __PTC_MMX__}
var
  p: PLongA;
  q: PSmallA;
  i: integer;
  r,g,b: smallInt;
{$endif}
begin
{$ifdef __PTC_MMX__}
//   mmx_convert_32_to_16_bgr555(dst, src, pixels);
{$else}
   p := src;
   q := dest;

   for i := pixels - 1 downto 0 do begin
     r := (p[0] and $00F80000) shr 20;
     g := (p[0] and $0000F800) shr 6;
     b := (p[0] and $000000F8) shl 8;

     q[0] := (r or g or b);
     Inc(p);
     Inc(q);
   end;
{$endif}
end;

function ptc_request_converter(bits, r, g,b: integer): TConvert;
begin
 result := nil;

   // 32bit RGB888 -> 32bit RGB888
  if (bits = 32) and (r = $00FF0000) and (g = $0000FF00) and (b = $000000FF) then
    result := ptc_convert_32_to_32_rgb888;

  // 32bit RGB888 -> 32bit BGR888
  if (bits = 32) and (r = $000000FF) and (g = $0000FF00) and (b = $00FF0000) then
    result := ptc_convert_32_to_32_bgr888;

  // 32bit RGB888 -> 24bit RGB888
  if (bits = 24) and (r = $00FF0000) and (g = $0000FF00) and (b = $000000FF) then
    result := ptc_convert_32_to_24_rgb888;

  // 32bit RGB888 -> 24bit BGR888
  if (bits = 24) and (r = $000000FF) and (g = $0000FF00) and (b = $00FF0000) then
    result := ptc_convert_32_to_24_bgr888;

  // 32bit RGB888 -> 16bit RGB565
  if (bits = 16) and (r = $F800) and (g = $07E0) and (b = $001F) then
    result := ptc_convert_32_to_16_rgb565;

  // 32bit RGB888 -> 16bit BGR565
  if (bits = 16) and (r = $001F) and (g = $07E0) and (b = $F800) then
    result := ptc_convert_32_to_16_bgr565;

  // 32bit RGB888 -> 16bit RGB555
  if (bits = 16) and (r = $7C00) and (g = $03E0) and (b = $001F) then
    result := ptc_convert_32_to_16_rgb555;

  // 32bit RGB888 -> 16bit BGR555
  if (bits = 16) and (r = $001F) and (g = $03E0) and (b = $7C00) then
    result := ptc_convert_32_to_16_bgr555;
end;

end.
