unit ddraw7;

interface

uses
  windows;

function  ptc_open(title: string; width, height: Integer): Integer;
function  ptc_update(buffer: Pointer): Integer;
procedure ptc_close;

var
  wnd: HWND;

implementation
{$ifdef __PTC_ICON__}
{$R 'tinyptc.res'}
{$endif}

uses
  Messages, convert, directdraw;

const
  __PTC_ICON__: string = 'IDI_MAIN';
  // menu option identifier
  SC_ZOOM_MSK: Integer = $400;
  SC_ZOOM_1x1: Integer = $401;
  SC_ZOOM_2x2: Integer = $402;
  SC_ZOOM_4x4: Integer = $404;

type
  TDirectDrawCreate = function (lpGUID: PGUID;
                                 out lplpDD: IDirectDraw;
                                 pUnkOuter: IUnknown) : HResult; stdcall;

var
  wc: WNDCLASS;
  dx,dy:            Integer;
  libddraw:         HMODULE;
  lpDD:             IDIRECTDRAW7;
  lpDDS:            IDIRECTDRAWSURFACE7;
  lpDDS_back:       IDIRECTDRAWSURFACE7;
  converter:        TConvert;
{$ifdef __PTC_WINDOWED__}
  lpDDC: IDIRECTDRAWCLIPPER;
  lpDDS_secondary: IDIRECTDRAWSURFACE7;
{$ifdef __PTC_CENTER_WINDOW__}
  original_window_width: Integer;
  original_window_height: Integer;
  system_menu: HMENU;
{$endif}
{$else}
  active:           Boolean;
{$endif}

{$ifdef __PTC_WINDOWED__}
procedure ptc_paint_primary;
var
  source: TRect;
  destination: TRect;
  point: TPoint;
begin
  if not assigned(lpDDS) then
    Exit;

  // setup source rectangle
  source.left := 0;
  source.top  := 0;
  source.right  := dx;
  source.bottom := dy;

  // get origin of client area
  point.x := 0;
  point.y := 0;
  ClientToScreen(wnd, point);

  // get window client area
  GetClientRect(wnd, destination);

  // offset destination rectangle
  destination.left   := destination.left + point.x;
  destination.top    := destination.top + point.y;
  destination.right  := destination.right + point.x;
  destination.bottom := destination.bottom + point.y;

  // blt secondary to primary surface
  lpDDS.Blt(@destination, lpDDS_secondary, @source, DDBLT_WAIT, nil);
end;
{$endif}

function WndProc(HWindow: HWnd; Message, WParam, LParam: Longint): Longint; stdcall;
{$ifdef __PTC_WINDOWED__}
{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
var
  zoom: Integer;
{$ifdef __PTC_CENTER_WINDOW__}
  x,y: integer;
{$endif}
{$endif}
{$endif}
{$endif}
begin
  // result data
  Result := 0;

	case (message) of
{$ifdef __PTC_WINDOWED__}
		WM_PAINT:
      begin
        // paint primary
        ptc_paint_primary();

        // call default window painting
        result := DefWindowProc(HWindow, message, wParam, lParam);
      end;
{$else}
    WM_ACTIVATEAPP:
      // update active flag
      active := wParam = 0;
    WM_SETCURSOR:
      // hide cursor
      SetCursor(0);
{$endif}

{$ifdef __PTC_WINDOWED__}
{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
		// check for message from our system menu entry
		WM_SYSCOMMAND:
			if ((wParam and $FFFFFFF0) = LongWord(SC_ZOOM_MSK)) then begin
{$ifdef __PTC_CENTER_WINDOW__}
				zoom := wParam and $7;
			  x := (GetSystemMetrics(SM_CXSCREEN) - original_window_width*zoom) shr 1;
				y := (GetSystemMetrics(SM_CYSCREEN) - original_window_height*zoom) shr 1;
				SetWindowPos(HWindow, 0, x, y, original_window_width * zoom,
                     original_window_height * zoom, SWP_NOZORDER);
{$else}
				zoom := wParam and $7;
  			SetWindowPos(HWindow, 0, 0, 0, original_window_width * zoom,
                      original_window_height * zoom, SWP_NOMOVE or SWP_NOZORDER);
{$endif}
      end else begin
        Result := DefWindowProc(HWindow, message, wParam, lParam)
      end;
{$endif}
{$endif}
{$endif}
{$ifdef __PTC_CLOSE_ON_ESCAPE__}
    WM_KEYDOWN:
      // close on escape key
      if ((wParam and $FF) = 27) then
        SendMessage(HWindow, WM_CLOSE, 0, 0);
{$endif}
{$ifdef __PTC_ALLOW_CLOSE__}
    WM_CLOSE:
      begin
        // close ptc
		    ptc_close();

	      // exit process
		    ExitProcess(0);
      end;
{$endif}
  	else
  		// unhandled messages
			Result := DefWindowProc(HWindow, message, wParam, lParam);
  end;
end;

function ptc_open(title: string; width, height: Integer): Integer;
var
{$ifdef __PTC_WINDOWED__}
  x,y : Integer;
  rect: TRect;
{$else}
  capabilities: TDDSCAPS2;
{$endif}
  format: TDDPixelFormat;
  descriptor: TDDSurfaceDesc2;
begin
  Result := 0;

  // setup data
  dx := width;
  dy := height;

  // load direct draw library
	libddraw := LoadLibrary('ddraw.dll');
  if (libddraw = 0) then
     Exit;

  // get directdraw create function address
  @DirectDrawCreateEx := GetProcAddress(libddraw ,'DirectDrawCreateEx');
	if (@DirectDrawCreateEx = nil) then
    Exit;

  if DirectDrawCreateEx(nil, lpDD, IDirectDraw7, nil) <> DD_OK then
    Exit;

{$ifndef __PTC_WINDOWED__}
	// register window class
	wc.style := 0;
	wc.lpfnWndProc := @WndProc;
	wc.cbClsExtra := 0;
	wc.cbWndExtra := 0;
{$ifdef __PTC_ICON__}
	wc.hInstance := GetModuleHandle(nil);
	wc.hIcon := LoadIcon(wc.hInstance, PChar(__PTC_ICON__));
{$else}
	wc.hInstance := 0;
	wc.hIcon := 0;
{$endif}
	wc.hCursor := LoadCursor(0, IDC_ARROW);
	wc.hbrBackground := 0;
	wc.lpszMenuName := nil;
	wc.lpszClassName := PChar(Title);
	if Windows.RegisterClass(wc) = 0 then
    Exit;

    // create window
{$ifdef __PTC_ICON__}
  wnd := CreateWindow(PChar(title), PChar(title), WS_POPUP or WS_SYSMENU,
                        0, 0, 0, 0, 0, 0, 0, nil);
{$else}
  wnd := CreateWindow(PChar(title), PChar(title),WS_POPUP,
                        0, 0, 0, 0, 0, 0, 0, nil);
{$endif}

  // enter exclusive mode
  if lpDD.SetCooperativeLevel(wnd, DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN) <> DD_OK then
     Exit;

  // enter display mode
  if lpDD.SetDisplayMode(width, height, 32,0,0) <> DD_OK then
    if lpDD.SetDisplayMode(width, height, 24,0,0) <> DD_OK then
      if lpDD.SetDisplayMode(width, height, 16,0,0) <> DD_OK then
        Exit;

  // primary with two back buffers
  FillChar(descriptor, sizeof(descriptor), 0);
  descriptor.dwSize  := sizeof(descriptor);
  descriptor.dwFlags := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
  descriptor.dwBackBufferCount := 2;
  descriptor.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_VIDEOMEMORY or
                                DDSCAPS_COMPLEX or DDSCAPS_FLIP;

  if lpDD.CreateSurface(descriptor, lpDDS, nil) <> DD_OK then begin
    // primary with one back buffer
    descriptor.dwBackBufferCount := 1;
    if lpDD.CreateSurface(descriptor, lpDDS, nil) <> DD_OK then begin
      // primary with no back buffers
      descriptor.dwFlags := DDSD_CAPS;
      descriptor.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_VIDEOMEMORY;
      if lpDD.CreateSurface(descriptor, lpDDS, nil) <> DD_OK then
        Exit;
    end;
  end;

  // get back surface
  capabilities.dwCaps := DDSCAPS_BACKBUFFER;
  if lpDDS.GetAttachedSurface(capabilities, lpDDS_back) <> DD_OK then
    Exit;

{$else}
	wc.style := 0;
	wc.lpfnWndProc := @WndProc;
	wc.cbClsExtra := 0;
	wc.cbWndExtra := 0;
{$ifdef __PTC_ICON__}
	wc.hInstance := GetModuleHandle(nil);
	wc.hIcon := LoadIcon(wc.hInstance, PChar(__PTC_ICON__));
{$else}
	wc.hInstance := 0;
	wc.hIcon := 0;
{$endif}
	wc.hCursor := LoadCursor(0, IDC_ARROW);
	wc.hbrBackground := 0;
	wc.lpszMenuName := nil;
	wc.lpszClassName := PChar(Title);
	if Windows.RegisterClass(wc) = 0 then
    Exit;

	// calculate window size
	rect.left   := 0;
	rect.top    := 0;
	rect.right  := width;
	rect.bottom := height;
	AdjustWindowRect(rect, WS_OVERLAPPEDWINDOW, False);
	rect.right  := rect.right - rect.left;
	rect.bottom := rect.bottom - rect.top;

{$ifdef __PTC_CENTER_WINDOW__}
	// center window
	x := (GetSystemMetrics(SM_CXSCREEN) - rect.right) shr 1;
	y := (GetSystemMetrics(SM_CYSCREEN) - rect.bottom) shr 1;
{$else}
	// let windows decide
	x := Integer(CW_USEDEFAULT);
	y := Integer(CW_USEDEFAULT);
{$endif}

{$ifdef __PTC_RESIZE_WINDOW__}
  // create resizable window
	wnd := CreateWindow(PChar(title), PChar(title), WS_OVERLAPPEDWINDOW,
                       x, y, rect.right, rect.bottom, 0, 0, 0, nil);
{$else}
  // create fixed window
	wnd := CreateWindow(PChar(title), PChar(title), 
                       WS_OVERLAPPEDWINDOW and (not WS_MAXIMIZEBOX) and (not WS_THICKFRAME),
                        x, y, rect.right, rect.bottom, 0, 0, 0, nil);
{$endif}

  // show window
  ShowWindow(wnd,SW_NORMAL);

{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
	// add entry to system menu to restore original window size
  system_menu := GetSystemMenu(wnd,FALSE);
	AppendMenu(system_menu, MF_STRING, SC_ZOOM_1x1, 'Zoom 1 x 1');
	AppendMenu(system_menu, MF_STRING, SC_ZOOM_2x2, 'Zoom 2 x 2');
	AppendMenu(system_menu, MF_STRING, SC_ZOOM_4x4, 'Zoom 4 x 4');

  // save original window size
  original_window_width := rect.right;
  original_window_height := rect.bottom;
{$endif}
{$endif}

  // enter cooperative mode
  if lpDD.SetCooperativeLevel(wnd, DDSCL_NORMAL) <> DD_OK then
    Exit;

  // primary with no back buffers
  FillChar(descriptor, sizeof(descriptor), 0);
  descriptor.dwSize  := sizeof(descriptor);
  descriptor.dwFlags := DDSD_CAPS;
  descriptor.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_VIDEOMEMORY;
  if lpDD.CreateSurface(descriptor, lpDDS, nil) <> DD_OK then
    Exit;

  // create secondary surface
  descriptor.dwFlags := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
  descriptor.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
  descriptor.dwWidth := width;
  descriptor.dwHeight := height;

  if lpDD.CreateSurface(descriptor, lpDDS_secondary, nil) <> DD_OK then
    Exit;

  // create clipper
  if lpDD.CreateClipper(0, lpDDC, nil) <> DD_OK then
    Exit;

  // set clipper to window
  if lpDDC.SetHWnd(0, wnd) <> DD_OK then
    Exit;

  // attach clipper object to primary surface
  if lpDDS.SetClipper(lpDDC) <> DD_OK then
    Exit;

  // set back to secondary
  lpDDS_back := lpDDS_secondary;
{$endif}

  // get pixel format
  format.dwSize := sizeof(format);
  if lpDDS.GetPixelFormat(format) <> DD_OK then
    Exit;

  // check that format is direct color
  if (format.dwFlags and DDPF_RGB) = 0 then
    Exit;

  // request converter function
  converter := ptc_request_converter(format.dwRGBBitCount, format.dwRBitMask,
                                    format.dwGBitMask, format.dwBBitMask);
  if not assigned(converter) then
    Exit;


{$ifdef __PTC_DISABLE_SCREENSAVER__}
    // disable screensaver while ptc is open
    SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 0, nil, 0);
{$endif}

  Result := 1;
end;

function ptc_update(buffer: Pointer): Integer;
type
  PByteA = ^TByteA;
  TByteA = array [0..0] of Byte;
var
  message: MSG;
  y: integer;
  descriptor: TDDSURFACEDESC2;
  src: PByteA;
  dst: PByteA;
  src_pitch: integer;
  dst_pitch: integer;
begin
  result := 0;

	// process messages
	while (PeekMessage(message, wnd, 0, 0, PM_REMOVE)) do begin
		// translate and dispatch
		TranslateMessage(message);
		DispatchMessage(message);
	end;

{$ifndef __PTC_WINDOWED__}
  if (active) then begin
{$endif}
    // restore surfaces
    lpDDS._Restore;
{$ifdef __PTC_WINDOWED__}
    lpDDS_secondary._Restore;
{$endif}

    // lock back surface
    descriptor.dwSize := sizeof(descriptor);
    if lpDDS_back.Lock(nil, descriptor, DDLOCK_SURFACEMEMORYPTR+DDLOCK_WAIT, 0) <> DD_OK then
      Exit;

    // calculate pitches
    src_pitch := dx * 4;
    dst_pitch := descriptor.lPitch;

    // copy pixels to back surface

    src := buffer;
    dst := descriptor.lpSurface;

    for y := 0 to dy - 1 do begin
      // convert line
      converter(src, dst, dx);
      Integer(src) := Integer(src) + src_pitch;
      Integer(dst) := Integer(dst) + dst_pitch;
    end;

    // unlock back surface
    lpDDS_back.Unlock(descriptor.lpSurface);

{$ifndef __PTC_WINDOWED__}
    // flip primary surface
    lpDDS.Flip(nil, DDFLIP_WAIT);
{$else}
    // paint primary
    ptc_paint_primary();
{$endif}

    // sleep
    Sleep(0);
{$ifndef __PTC_WINDOWED__}
  end else begin
    Sleep(1);
  end;
{$endif}

 // success
 result := 1;
end;

procedure ptc_close;
begin
{$ifdef __PTC_WINDOWED__}
  // check secondary
  if assigned(lpDDS_secondary) then begin
    // release secondary
    lpDDS_secondary := nil;
  end;
{$endif}

  // check
  if assigned(lpDDS) then begin
    // release primary
    lpDDS := nil;
  end;

  // check
  if assigned(lpDD) then begin
    // leave display mode
    lpDD.RestoreDisplayMode;

    // leave exclusive mode
    lpDD.SetCooperativeLevel(wnd, DDSCL_NORMAL);

    // free direct draw
    lpDD := nil;
  end;

  // hide window
  DestroyWindow(wnd);

{$ifdef __PTC_DISABLE_SCREENSAVER__}
  // enable screensaver now that ptc is closed
  SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 1, nil, 0);
{$endif}
end;

end.


