unit gdi;

interface

uses
  windows;

function  ptc_open(title: string; width, height: Integer): Integer;
function  ptc_update(buffer: Pointer): Integer;
procedure ptc_close;

var
  wnd: HWND;
  window_hdc: HDC;

implementation
{$ifdef __PTC_ICON__}
{$R 'tinyptc.res'}
{$endif}

uses
  Messages;

const
  __PTC_ICON__: string = 'IDI_MAIN';
  // menu option identifier
  SC_ZOOM_MSK: Integer = $400;
  SC_ZOOM_1x1: Integer = $401;
  SC_ZOOM_2x2: Integer = $402;
  SC_ZOOM_4x4: Integer = $404;

type
  HDRAWDIB  = THandle;
  PVOID     = Pointer;

var
  wc: WNDCLASS;
  surface_width: Integer;
  surface_height: Integer;
  original_window_x: Integer;
  original_window_y: Integer;
  original_window_width: Integer;
  original_window_height: Integer;
  cached_buffer: Pointer;
  bitmap_header: PbitmapInfoheader;

function WndProc(HWindow: HWnd; Message, WParam, LParam: Longint): Longint; stdcall;
{$ifdef __PTC_RESIZE_WINDOW__}
var
  windowsize: TRECT;
{$endif}
{$ifdef __PTC_WINDOWED__}
{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
var
  zoom: Integer;
{$ifdef __PTC_CENTER_WINDOW__}
  x,y: integer;
{$endif}
{$endif}
{$endif}
{$endif}
begin
  Result := 0;
	// handle message
	case (message) of
		WM_PAINT:
      // check cached buffer
      if Assigned(cached_buffer) then begin
{$ifdef __PTC_RESIZE_WINDOW__}
        // grab current window size
        GetClientRect(HWindow, windowsize);
        StretchDIBits(window_hdc, 0, 0, windowsize.right, windowsize.bottom,
   		                 0, 0, surface_width, surface_height, cached_buffer,
                        PBitmapInfo(bitmap_header)^, DIB_RGB_COLORS, SRCCOPY);
{$else}
        StretchDIBits(window_hdc, 0, 0, surface_width, surface_height, 0, 0,
                       surface_width, surface_height, cached_buffer,
                       PBitmapInfo(bitmap_header)^, DIB_RGB_COLORS, SRCCOPY);
{$endif}

        // validate window
        ValidateRect(wnd, nil);
      end;
{$ifdef __PTC_WINDOWED__}
{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
		// check for message from our system menu entry
		WM_SYSCOMMAND:
			if ((wParam and $FFFFFFF0) = LongWord(SC_ZOOM_MSK)) then begin
{$ifdef __PTC_CENTER_WINDOW__}
				zoom := wParam and $7;
			  x := (GetSystemMetrics(SM_CXSCREEN) - original_window_width*zoom) shr 1;
				y := (GetSystemMetrics(SM_CYSCREEN) - original_window_height*zoom) shr 1;
				SetWindowPos(HWindow, 0, x, y, original_window_width * zoom,
                     original_window_height * zoom, SWP_NOZORDER);
{$else}
				zoom := wParam and $7;
  			SetWindowPos(HWindow, 0, 0, 0, original_window_width * zoom,
                      original_window_height * zoom, SWP_NOMOVE or SWP_NOZORDER);
{$endif}
      end else begin
        Result := DefWindowProc(HWindow, message, wParam, lParam)
      end;
{$endif}
{$endif}
{$endif}
{$ifdef __PTC_CLOSE_ON_ESCAPE__}
    WM_KEYDOWN:
      // close on escape key
      if ((wParam and $FF) = 27) then
        SendMessage(HWindow, WM_CLOSE, 0, 0);
{$endif}
{$ifdef __PTC_ALLOW_CLOSE__}
    WM_CLOSE:
      begin
        // close ptc
		    ptc_close();

	      // exit process
		    ExitProcess(0);
      end;
{$endif}
  	else
  		// unhandled messages
			Result := DefWindowProc(HWindow, message, wParam, lParam);
  end;
end;

type
  PLongA = ^TLongA;
  TLongA = array [Byte] of Longint;

function ptc_open(title: string; width, height: Integer): Integer;
var
  rect: TRect;
  BitmapInfoSize: integer;
  Colors: PLongA;
{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
  menu: HMENU;
{$endif}
{$endif}
begin
  result :=  0;

	// register window class
	wc.style := CS_OWNDC or CS_VREDRAW or CS_HREDRAW;      // very important
	wc.lpfnWndProc := @WndProc;
	wc.cbClsExtra := 0;
	wc.cbWndExtra := 0;
{$ifdef __PTC_ICON__}
	wc.hInstance := GetModuleHandle(nil);
	wc.hIcon := LoadIcon(wc.hInstance, PChar(__PTC_ICON__));
{$else}
	wc.hInstance := 0;
	wc.hIcon := 0;
{$endif}
	wc.hCursor := LoadCursor(0, IDC_ARROW);
	wc.hbrBackground := 0;
	wc.lpszMenuName := nil;
	wc.lpszClassName := PChar(Title);
	if Windows.RegisterClass(wc) = 0 then
    Exit;

	// calculate window size
	rect.left   := 0;
	rect.top    := 0;
	rect.right  := width;
	rect.bottom := height;
{$ifdef __PTC_RESIZE_WINDOW__}
	AdjustWindowRect(rect, WS_OVERLAPPEDWINDOW, False);
{$else}
	AdjustWindowRect(rect, WS_POPUP or WS_SYSMENU or WS_CAPTION, False);
{$endif}
	rect.right  := rect.right - rect.left;
	rect.bottom := rect.bottom - rect.top;

	// save surface size and original window size
	surface_width := width;
	surface_height := height;
	original_window_width := rect.right;
	original_window_height := rect.bottom;

{$ifdef __PTC_CENTER_WINDOW__}
	// center window
	original_window_x := (GetSystemMetrics(SM_CXSCREEN) - rect.right) shr 1;
	original_window_y := (GetSystemMetrics(SM_CYSCREEN) - rect.bottom) shr 1;
{$else}
	// let windows decide
	original_window_x := Integer(CW_USEDEFAULT);
	original_window_y := Integer(CW_USEDEFAULT);
{$endif}

	// create window and show it
{$ifdef __PTC_RESIZE_WINDOW__}
	wnd := CreateWindowEx(0, PChar(title), PChar(title), WS_OVERLAPPEDWINDOW,
                         original_window_x, original_window_y, rect.right,
                          rect.bottom, 0, 0, 0, nil);

{$else}
	wnd := CreateWindowEx(0, PChar(title), PChar(title),
                         WS_OVERLAPPEDWINDOW and (not WS_MAXIMIZEBOX) and (not WS_THICKFRAME),
                          original_window_x, original_window_y, rect.right,
                           rect.bottom, 0, 0, 0, nil);
{$endif}

	ShowWindow(wnd, SW_NORMAL);

	// get window dc
	window_hdc := GetDC(wnd);

  // create bitmap header
//  for cc := 0 to sizeof(BITMAPINFOHEADER) + 16 - 1 do
//		bitmapbuffer[cc] := 0;
//  bitmap_header := @bitmapbuffer[0];
  BitmapInfoSize := sizeof(TBITMAPINFOHEADER) + 3 * SizeOf(TRGBQuad);
  GetMem(bitmap_header, BitmapInfoSize);

	bitmap_header^.biSize        := sizeof(TBITMAPINFOHEADER);
	bitmap_header^.biPlanes      := 1;
	bitmap_header^.biBitCount    := 32;
	bitmap_header^.biCompression := BI_BITFIELDS;
	bitmap_header^.biWidth       := surface_width;
	bitmap_header^.biHeight      := -surface_height;	// note well
  Colors := Pointer(DWORD(bitmap_header) + sizeof(TBITMAPINFOHEADER));
  Integer(Colors[0]) := $00FF0000;
  Integer(Colors[1]) := $0000FF00;
  Integer(Colors[2]) := $000000FF;

{$ifdef __PTC_RESIZE_WINDOW__}
{$ifdef __PTC_SYSTEM_MENU__}
		// add entry to system menu to restore original window size
		 menu := GetSystemMenu(wnd,FALSE);
		AppendMenu(menu, MF_STRING, SC_ZOOM_1x1, 'Zoom 1 x 1');
		AppendMenu(menu, MF_STRING, SC_ZOOM_2x2, 'Zoom 2 x 2');
		AppendMenu(menu, MF_STRING, SC_ZOOM_4x4, 'Zoom 4 x 4');
{$endif}
{$endif}

{$ifdef __PTC_DISABLE_SCREENSAVER__}
    // disable screensaver while ptc is open
    SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 0, nil, 0);
{$endif}

	// success
	result := 1;
end;

function ptc_update(buffer: Pointer): Integer;
var
  message: MSG;
begin
	// update buffer ptr cache
	cached_buffer := buffer;

	// invalidate window
	InvalidateRect(wnd, nil, TRUE);

	// send paint window message
	SendMessage(wnd, WM_PAINT, 0, 0);

	// process messages
	while (PeekMessage(message, wnd, 0, 0, PM_REMOVE)) do begin
		// translate and dispatch
		TranslateMessage(message);
		DispatchMessage(message);
	end;

	// sleep
	Sleep(0);

	// success
	result := 1;
end;

procedure ptc_close;
begin
	// clear cached buffer
	cached_buffer := nil;

	// release DC
	ReleaseDC(wnd, window_hdc);

	// destroy window
	DestroyWindow(wnd);

{$ifdef __PTC_DISABLE_SCREENSAVER__}
  // enable screensaver now that ptc is closed
  SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 1, nil, 0);
{$endif}
end;


end.
